import tensorflow as tf
from sklearn.datasets import load_digits
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder
import numpy as np

config = tf.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.Session(config=config)


class CNNRegressor:
    
    """
    初始化的输入参数convo_para是卷积池化层的参数,里面的每一个值，
    是一个列表[con_fiter,con_strides,pool_filter,pool_strides]
    其中的filter是[channels,heights,widths,deepth],strides是向各方向移动的步长
    """

    def __init__(self, convo_para=[
        [[10, 3, 3, 1], [1, 1, 1, 1], [1, 3, 3, 1], [1, 2, 2, 1]],
        [[5, 3, 3, 1], [1, 1, 1, 1], [1, 3, 3, 1], [1, 2, 2, 1]],
    ], hidden_layer_size=50, max_iter=100,times=0.2):
        self.batch_size = 8
        self.middle_layers_size = len(convo_para)
        self.conv_filters_sizes = [convo_para[i][0] for i in range(self.middle_layers_size)]
        self.conv_strides = [convo_para[i][1] for i in range(self.middle_layers_size)]
        self.pool_filters_sizes = [convo_para[i][2] for i in range(self.middle_layers_size)]
        self.pool_strides = [convo_para[i][3] for i in range(self.middle_layers_size)]
        self.hidden_layer_size = hidden_layer_size
        self.max_iter = max_iter
        self.times = times

    def generatebatch(self, X, Y, n_examples, batch_size):
        for batch_i in range(n_examples // batch_size):
            start = batch_i*batch_size
            end = start + batch_size
            batch_xs = X[start:end]
            batch_ys = Y[start:end]
            yield batch_xs, batch_ys  # 生成每一个batch

    def _fit(self, X, Y):
        """
        :param X: X是（batch, height, width, channels）数据，n是样本个数
        :param Y: Y是(batch, results)数据
        :return:
        """
        batch, input_height, input_width, input_deepth = X.shape
        tf.reset_default_graph()
        # 输入层
        tf_X = tf.placeholder(tf.float32, [None, input_height, input_width,  input_deepth])
        tf_Y = tf.placeholder(tf.float32, [None, Y.shape[1]])

        self.output_size = Y.shape[1]

        conv_X = tf_X

        self.conv_filter_w, self.conv_filter_b = [], []
        self.batch_mean, self.batch_var = [], []
        self.shift, self.scale = [], []

        for i in range(self.middle_layers_size):
            if i != 0:
                w_size = self.conv_filters_sizes[i][1:3] + [self.conv_filters_sizes[i-1][0]] + [self.conv_filters_sizes[i][0]]
            else:
                w_size = self.conv_filters_sizes[i][1:] + [self.conv_filters_sizes[i][0]]
            self.conv_filter_w.append(tf.Variable(tf.random_normal(w_size)))
            self.conv_filter_b.append(tf.Variable(tf.random_normal([self.conv_filters_sizes[i][0]])))

            # 卷积层
            conv_out = tf.nn.conv2d(conv_X, self.conv_filter_w[i], strides=self.conv_strides[i], padding='SAME') + self.conv_filter_b[i]

            # BN层
            if i!=0:
                batch_mean, batch_var = tf.nn.moments(conv_out, [0, 1, 2], keep_dims=True)
                self.batch_mean.append(batch_mean)
                self.batch_var.append(batch_var)
                self.shift.append(tf.Variable(tf.zeros([self.conv_filters_sizes[i][0]])))
                self.scale.append(tf.Variable(tf.ones([self.conv_filters_sizes[i][0]])))
                epsilon = 1e-3
                BN_out = tf.nn.batch_normalization(conv_out, self.batch_mean[i-1], self.batch_var[i-1], self.shift[i-1], self.scale[i-1], epsilon)
            else:
                BN_out = conv_out

            # 激活层
            relu_feature_maps = tf.nn.relu(BN_out)

            # 池化层
            max_pool = tf.nn.max_pool(relu_feature_maps, ksize=self.pool_filters_sizes[i], strides=self.pool_strides[i], padding='SAME')

            conv_X = max_pool

        input_size = max_pool.shape[1]*max_pool.shape[2]*max_pool.shape[3]
        max_pool_flat = tf.reshape(max_pool, [-1, input_size])

        # 全连接层
        self.fc_w = tf.Variable(tf.random_normal([input_size, self.hidden_layer_size]))
        self.fc_b = tf.Variable(tf.random_normal([self.hidden_layer_size]))
        fc_out1 = tf.nn.relu(tf.matmul(max_pool_flat, self.fc_w) + self.fc_b)

        # 输出层
        self.out_w = tf.Variable(tf.random_normal([self.hidden_layer_size, self.output_size]))
        self.out_b = tf.Variable(tf.random_normal([self.output_size]))
        pred = tf.nn.softmax(tf.matmul(fc_out1, self.out_w)+self.out_b)

        loss = -tf.reduce_mean(tf_Y*tf.log(tf.clip_by_value(pred, 1e-11, 1.0)))
        train_step = tf.train.AdamOptimizer(1e-3).minimize(loss)

        with tf.Session() as sess:
            sess.run(tf.global_variables_initializer())
            for epoch in range(self.max_iter):
                for batch_xs, batch_ys in self.generatebatch(X, Y, Y.shape[0], self.batch_size): # 每个周期进行MBGD算法
                    sess.run(train_step, feed_dict={tf_X: batch_xs, tf_Y: batch_ys})


            # 运算出神经网络参数
            for i in range(self.middle_layers_size):
                self.conv_filter_w[i] = sess.run(self.conv_filter_w[i])
                self.conv_filter_b[i] = sess.run(self.conv_filter_b[i])
                if i != 0:
                    self.batch_mean[i - 1] = sess.run(self.batch_mean[i - 1], feed_dict={tf_X: X, tf_Y: Y})
                    self.batch_var[i - 1] = sess.run(self.batch_var[i - 1], feed_dict={tf_X: X, tf_Y: Y})
                    self.shift[i - 1] = sess.run(self.shift[i - 1], feed_dict={tf_X: X, tf_Y: Y})
                    self.scale[i - 1] = sess.run(self.scale[i - 1], feed_dict={tf_X: X, tf_Y: Y})
            self.fc_w, self.fc_b = sess.run(self.fc_w), sess.run(self.fc_b)
            self.out_w, self.out_b = sess.run(self.out_w), sess.run(self.out_b)


    def _predict(self, X):

        tf.reset_default_graph()
        x = tf.placeholder(tf.float32, X.shape)
        conv_X = x
        for i in range(self.middle_layers_size):
            # 卷积层
            conv_out = tf.nn.conv2d(conv_X, self.conv_filter_w[i], strides=self.conv_strides[i], padding='SAME') + \
                       self.conv_filter_b[i]

            # BN层
            if i != 0:
                epsilon = 1e-3
                BN_out = tf.nn.batch_normalization(conv_out, self.batch_mean[i-1], self.batch_var[i-1], self.shift[i-1], self.scale[i-1], epsilon)
            else:
                BN_out = conv_out

            # 激活层
            relu_feature_maps = tf.nn.relu(BN_out)

            # 池化层
            max_pool = tf.nn.max_pool(relu_feature_maps, ksize=self.pool_filters_sizes[i], strides=self.pool_strides[i],
                                      padding='SAME')

            conv_X = max_pool

        input_size = max_pool.shape[1] * max_pool.shape[2] * max_pool.shape[3]
        max_pool_flat = tf.reshape(max_pool, [-1, input_size])

        # 全连接层
        fc_out1 = tf.nn.relu(tf.matmul(max_pool_flat, self.fc_w) + self.fc_b)

        # 输出层
        pred = tf.nn.softmax(tf.matmul(fc_out1, self.out_w) + self.out_b)

        with tf.Session() as sess:
            sess.run(tf.global_variables_initializer())
            output = sess.run(pred, feed_dict={x: X})

        return output


    def fit(self, X, Y):
        """
        :param X: X是（n, height, width, deepth）数据，n是样本个数
        :param Y: Y是(n, results)数据
        :return:
        """
        self.x_shape = X.shape
        std_x, self.scale_x = self.standard(X.reshape([self.x_shape[0],-1]))
        std_y, self.scale_y = self.standard(Y)
        self._fit(std_x.reshape(self.x_shape), std_y)


    def predict(self, X):
        shape = [-1] + list(self.x_shape[1:])
        std_x = self.standard(X.reshape([-1,shape[1]*shape[2]*shape[3]]), scale=self.scale_x )
        output = self._predict(std_x.reshape(shape))
        pre = self.unstandard(output, self.scale_y)
        return pre

    def score(self, X, Y):
        pred_y = self.predict(X)
        if len(Y.shape) == 1:
            pred_y.shape = Y.shape
        u = np.sum(np.power(pred_y - Y, 2))
        v = np.sum(np.power(pred_y - np.mean(pred_y, axis=0), 2))
        return 1 - u / v

    def standard(self, data, scale=0):
        if scale==0:
            a = np.min(data, axis=0)
            b = np.max(data, axis=0)-a+1e-10
            a = a - self.times*b
            b = (2*self.times+1)*b
            std_data=(data-a)/b
            scale_data=[a,b]
            return std_data,scale_data
        else:
            std_data=(data-scale[0])/(scale[1])
            return std_data

    def unstandard(self,std_data,scale_data):
        a=scale_data[0]
        b=scale_data[1]
        data = std_data*b+a
        return data


class CNNClassifier(CNNRegressor):
    def __init__(self, convo_para=[
        [[10, 3, 3, 1], [1, 1, 1, 1], [1, 3, 3, 1], [1, 2, 2, 1]],
        [[5, 3, 3, 1], [1, 1, 1, 1], [1, 3, 3, 1], [1, 2, 2, 1]],
    ], hidden_layer_size=50, max_iter=100,times=0.2):
        super().__init__(convo_para=convo_para, hidden_layer_size=hidden_layer_size, max_iter=max_iter, times=times)

    def fit(self, X, Y):
        self.x_shape = X.shape
        y = OneHotEncoder().fit_transform(Y).todense()
        std_x, self.scale_x = self.standard(X.reshape([self.x_shape[0],-1]))
        self._fit(std_x.reshape(self.x_shape), y)


    def predict(self, X):
        shape = [-1] + list(self.x_shape[1:])
        std_x = self.standard(X.reshape([-1,shape[1]*shape[2]*shape[3]]), self.scale_x)
        pro = self._predict(std_x.reshape(shape))
        pre = np.argmax(pro, axis=1)
        return pre

    def score(self, X, Y):
        pre = self.predict(X)
        if (Y.ndim==2) and (1 in Y.shape):
            real_y = Y.reshape(-1)
        else:
           real_y = Y
        right = np.equal(pre, real_y).sum()
        return right / len(Y)



if __name__ == '__main__':
    digits = load_digits()
    X_data = digits.data.astype(np.float32)
    Y_data = digits.target.astype(np.float32).reshape(-1, 1)
    # Y = OneHotEncoder().fit_transform(Y_data) # one-hot编码
    X = X_data.reshape(-1, 8, 8, 1)
    cnn = CNNClassifier(max_iter=100)
    cnn.fit(X, Y_data)
    print(cnn.score(X, Y_data))



import numpy as np
from random import randint
from sklearn.neural_network import MLPRegressor, MLPClassifier
from sklearn.preprocessing import OneHotEncoder
import warnings

class GA_MLPRegressor:

    def __init__(self, hidden_layer_sizes=[10], activation='relu', times=1, pop=5, max_iter=500):
        if type(hidden_layer_sizes) == list:
            self.hidden_layer_sizes = hidden_layer_sizes
        elif type(hidden_layer_sizes) == int:
            self.hidden_layer_sizes = [hidden_layer_sizes]
        self.activation = activation
        self.pop = pop
        self.max_iter = max_iter
        self.times = times
        if activation == 'logistic':
            self.method = '0_1'
        elif activation == 'relu':
            self.method = '0_inf'

    def _fit(self, x, y, class_weights=[1]):
        if (len(class_weights)>1) and (len(class_weights)<y.shape[1]):
            raise TypeError('参数class_weights长度与y的类别不匹配')
        try:
            std_x, self.scale_x = self.standard(x, method = self.method)
            std_y, self.scale_y = self.standard(y ,method = self.method)
            self.layers = [std_x.shape[1]] + self.hidden_layer_sizes + [std_y.shape[1] if std_y.ndim ==2 else 1]
            # 初始化种群
            population_w,population_b = list(),list()
            for j in range(self.pop):
                one_w,one_b=list(),list()      
                for i in range(len(self.layers)-1):
                    one_w.append(np.random.uniform(-1, 1, [self.layers[i], self.layers[i+1]]))
                    one_b.append(np.random.uniform(-1, 1, [1,self.layers[i+1] ]))
                population_w.append(one_w)
                population_b.append(one_b)
            # 开始遗传优化
            best_pop = [np.inf,0,0]
            iteration = 0
            while iteration <= self.max_iter:
                new_pop_w,new_pop_b=list(),list()
                # 计算适应度并记录最好个体
                pop_fitness=list()
                for i in range(len(population_w)):
                    pop_fitness.append(self.fitness(std_x,std_y,population_w[i],population_b[i],class_weights=class_weights))
                if min(pop_fitness) < best_pop[0]:
                    n=pop_fitness.index(min(pop_fitness))
                    best_pop = [pop_fitness[n] , population_w[n] , population_b[n]]
                for i in range(int(self.pop/2)):
                    # 选择操作
                    x,y=self.select(pop_fitness)
                    # 将个体建设为染色体
                    person1_w,person2_w = self.matrix_to_vector(population_w[x]), self.matrix_to_vector(population_w[y])
                    person1_b,person2_b = self.matrix_to_vector(population_b[x]), self.matrix_to_vector(population_b[y])
                    # 交叉互换
                    person1_w, person2_w, person1_b, person2_b = self.cross(person1_w, person2_w, person1_b, person2_b)
                    # 变异
                    person1_w=self.mut(person1_w)
                    person2_w=self.mut(person2_w)
                    person1_b=self.mut(person1_b)
                    person2_b=self.mut(person2_b)
                    # 将染色体进化为个体
                    person1_w = self.vector_to_weights(person1_w)
                    person2_w = self.vector_to_weights(person2_w)
                    person1_b = self.vector_to_bias(person1_b)
                    person2_b = self.vector_to_bias(person2_b)
                    # 构建新的种群
                    new_pop_w += [person1_w,person2_w]
                    new_pop_b += [person1_b,person2_b]
                new_pop_w.append(best_pop[1])
                new_pop_b.append(best_pop[2])
                population_w,population_b=new_pop_w,new_pop_b
                iteration += 1
                if iteration%100==0:
                    print('({}/{})complete'.format(iteration,self.max_iter))
                
            else:
                # 遗传进化完成，计算适应度并记录最好个体
                pop_fitness = [self.fitness(std_x, std_y, population_w[i], population_b[i],class_weights=class_weights) for i in
                               range(len(population_w))]
                if min(pop_fitness) < best_pop[0]:
                    n = pop_fitness.index(min(pop_fitness))
                    best_pop = [pop_fitness[n], population_w[n], population_b[n]]
                self.w, self.b = best_pop[1], best_pop[2]
        except KeyboardInterrupt:
            pop_fitness = [self.fitness(std_x, std_y, population_w[i], population_b[i]) for i in
                           range(len(population_w))]
            if min(pop_fitness) < best_pop[0]:
                n = pop_fitness.index(min(pop_fitness))
                best_pop = [pop_fitness[n], population_w[n], population_b[n]]
            self.w, self.b = best_pop[1], best_pop[2]
            warnings.warn('Training interrupted by user.')


    def fit(self, x, y):
        self._fit(x,y)

    def _predict(self, x):
        std_x = self.standard(x, method = self.method, scale = self.scale_x )
        out_puts=self.activate(np.dot(std_x, self.w[0]) - self.b[0])
        for i in range(len(self.layers)-2):
            out_puts=self.activate(np.dot(out_puts, self.w[i+1]) - self.b[i+1])
        pre = self.unstandard(out_puts, self.scale_y)
        if 1 in pre.shape:
            pre.shape = -1
        return pre

    def predict(self,x):
        return self._predict(x)


    def activate(self,x):
        if self.activation == 'logistic':
            return  (1 / (1 + np.exp(-x)))
        elif self.activation == 'relu':
            y=x.copy()
            y[y<0]=0
            return y

    def score(self,x,y):
        pred_y = self.predict(x)
        if len(y.shape)==1:
            pred_y.shape = y.shape
        u = np.sum(np.power(pred_y-y,2))
        v = np.sum(np.power(pred_y-np.mean(pred_y,axis=0),2))
        return 1-u/v

    #数据处理的一些操作所需函数
        
    def standard(self, data, method='0_1', scale=0):
        #数据标准化，返回标准化后的数据和数据的规格，该规格是2行的n列的矩阵列表，第一行表示减去的值，第二行表示长度 
        if scale==0:
            a = np.min(data, axis=0)
            b = np.max(data, axis=0)-a
            a = a - self.times*b
            b = (2*self.times+1)*b
            if method == '0_1':
                pass
            elif method == '0_inf':
                b=1
            std_data=(data-a)/b
            scale_data=[a,b]
            return std_data,scale_data           
        else:
            std_data=(data-scale[0])/scale[1]
            return std_data


    def unstandard(self,std_data,scale_data):
        a=scale_data[0]
        b=scale_data[1]
        data = std_data*b+a
        return data
        
    def matrix_to_vector(self,one):
        vec=list()
        for i in range(len(one)):
            vec+=(one[i].reshape(1,-1).tolist()[0])
        return vec

    def vector_to_weights(self,vec):
        mat=list()
        v=vec.copy()
        for i in range(len(self.layers)-1):
            num=self.layers[i]*self.layers[i+1]
            tem=np.array(v[0:num])
            tem.shape=self.layers[i],self.layers[i+1]
            del v[0:num]
            mat.append(tem)
        return mat

    def vector_to_bias(self,vec):
        bia=list()
        v=vec.copy()
        for i in range(len(self.layers)-1):
            bia.append(np.array(v[0:self.layers[i+1]]))
            del v[0:self.layers[i+1]]
        return bia
        
        

    #遗传算法的一些操作所需函数

    def fitness(self,x,y,w,b,class_weights):
        out_puts = self.activate(np.dot(x, w[0]) - b[0])
        for i in range(len(self.layers) - 2):
            out_puts = self.activate(np.dot(out_puts, w[i + 1]) - b[i + 1])
        pre = out_puts
        if len(y.shape)==1:
            pre.shape = y.shape
        ans = np.abs(pre-y)*np.array(class_weights)
        dist =np.sqrt(np.sum(np.power(ans,2)))
        return dist
    
    def select(self,f,d=15):
        #根据适应度选择两个个体，d为选择程度，适应度越小越容易被选中
        n=len(f)
        f=np.array(f)
        p=pow(1/f,d)
        p=p/sum(p)
        #以上求倒数乘方并归一化得出选取概率，以下产生随机数选取
        ra=np.random.random()
        s=0  #累积概率
        for i in range(n):
            s+=p[i]
            if ra<=s:
                x=i
                break
        p[x]=0
        p=p/sum(p)
        ra=np.random.random()
        s=0
        for i in range(n):
            s+=p[i]
            if ra<=s:
                y=i
                break
        return x,y

    def cross(self,x_1,y_1,x_2,y_2):
        n=len(x_1)
        #随机产生交叉位
        a,b=randint(0,n-1),randint(0,n-1)
        [a,b]=np.sort([a,b])
        (x_1[a:b+1],y_1[a:b+1])=(y_1[a:b+1],x_1[a:b+1])
        (x_2[a:b + 1], y_2[a:b + 1]) = (y_2[a:b + 1], x_2[a:b + 1])
        return x_1,y_1,x_2,y_2

    def mut(self,x):
        n=len(x)
        #随机产生变异位
        a,b=randint(0,n-1),randint(0,n-1)
        [a,b]=np.sort([a,b])
        random_value = np.random.uniform(-1,1,[1,b-a+1])
        x[a:b+1] = (np.array(x[a:b+1]) + random_value).tolist()[0]
        return x


class GA_MLPClassifier(GA_MLPRegressor):
    def __init__(self, hidden_layer_sizes=[10], activation='logistic', times=1, pop=5, max_iter=500):
        super().__init__(hidden_layer_sizes=hidden_layer_sizes, activation=activation, times=times, pop=pop, max_iter=max_iter)

    def fit(self, x, y, class_weights=[1]):
        y = np.array(y).reshape(-1,1)
        self.ohe = OneHotEncoder(sparse=False)
        y_dist = self.ohe.fit_transform(y)        
        self._fit(x, y_dist,class_weights=class_weights)

    def predict_proba(self,x):
        raw_pro = self._predict(x)
        raw_pro = raw_pro - np.min(raw_pro,axis=1).reshape(-1,1)
        s = np.sum(raw_pro,axis=1)
        pro = raw_pro/s.reshape(-1,1)
        return pro

    def predict(self, x):
        pro = self._predict(x)        
        pre = self.ohe.inverse_transform(pro)
        return pre.reshape(-1)

    def score(self,x,y):
        pre_y = self.predict(x)
        right=0
        n=len(pre_y)
        for i in range(n):
            if pre_y[i] == y[i]:
                right += 1
        return right/n






    
if __name__=='__main__':
    view = '分类回归分析'
    if view == '数值回归分析':
        #数值回归分析
        s1=GA_MLPRegressor(hidden_layer_sizes=10,max_iter=800,times=1.5,activation='relu')
        s2=MLPRegressor(hidden_layer_sizes=8,max_iter=800,activation='relu')
        #y=x1+2lnx2-3x3
        x=np.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12],[13,14,15],[16,17,18]])
        y=x[:,0]+2*np.log(x[:,1])-3*x[:,2]
        y.shape=-1,1
        test_x=np.array([[19,20,21],[22,23,24],[25,26,27]])
        test_y=test_x[:,0]+2*np.log(test_x[:,1])-3*test_x[:,2]
        test_y.shape=-1,1
        s1.fit(x,y)
        s2.fit(x,y)
        pre1_y=s1.predict(test_x)
        pre2_y=s2.predict(test_x)
        print('实际值：')
        print(test_y)
        print(' ')
        print('遗传改进网络输出和得分：')
        print(pre1_y)
        print(s1.score(test_x,test_y))
        print(' ')
        print('多层感知器输出和得分：')
        print(pre2_y)
        print(s2.score(test_x,test_y))
    elif  view == '分类回归分析':
        #分类回归分析
        s1 = GA_MLPClassifier(hidden_layer_sizes=8,max_iter=200,times=0.05,activation='logistic')
        s2 = MLPClassifier(hidden_layer_sizes=8,max_iter=400,activation='relu')
        x = np.array([[10, 2, 3], [4, 15, 6], [7, 8, 11], [1, 9, 12], [5, 14, 5], [16, 7, 8]])
        y = np.array(['a','b','c','c','b','a'])
        y.shape = -1, 1
        #x第一列超过10就是a,2列超过是b，3列超过是c
        test_x = np.array([[1, 14, 2], [12, 3, 4], [5, 6, 17]])
        test_y = np.array(['b','a','c'])
        test_y.shape = -1, 1
        s1.fit(x, y,[1,9,9])
        s2.fit(x, y)
        pre1_y = s1.predict(test_x)
        pre2_y = s2.predict(test_x)
        pro1 = s1.predict_proba(test_x)
        pro2 = s2.predict_proba(test_x)
        print('遗传改进网络输出概率和分类以及准确率：')
        print(pro1)
        print(pre1_y)
        print(s1.score(test_x, test_y))
        print('实际分类')
        print(test_y)
        print('多层感知器输出概率和分类以及准确率：')
        print(pro2)
        print(pre2_y)
        print(s2.score(test_x, test_y))

    


        
        
        
        



import numpy as np
from random import randint
from sklearn.neural_network import MLPRegressor, MLPClassifier
from sklearn.linear_model import LinearRegression
from sklearn.cluster import KMeans
from sklearn.preprocessing import OneHotEncoder

class RBFRegressor:
    def __init__(self, hidden_layer_sizes=10,  times=0.5, fit_intercept=False):
        self.hidden_layer_sizes = hidden_layer_sizes
        self.times = times
        self.fit_intercept = fit_intercept

    def _fit(self, x, y):
        std_x, self.scale_x = self.standard(x)
        std_y, self.scale_y = self.standard(y)
        if len(y.shape) == 1:
            output_layer_sizes = 1
        else:
            output_layer_sizes = y.shape[1]
        km = KMeans(n_clusters=self.hidden_layer_sizes)
        km.fit(std_x)
        hidden_nodes_c = [i for i in km.cluster_centers_]
        hidden_nodes_w = [np.mean(np.sqrt(np.sum(np.power(std_x-i, 2), axis=1))) for i in km.cluster_centers_]
        hidden_outputs = [np.exp(-np.sqrt(np.sum(np.power(std_x-hidden_nodes_c[i], 2), axis=1))/(np.power(hidden_nodes_w[i],2))) for i in range(self.hidden_layer_sizes)]
        hidden_outputs = np.array(hidden_outputs).T
        weights, intercepts = list(), [0 for i in range(output_layer_sizes)]
        for i in range(output_layer_sizes):
            if y.ndim==1:
                y_for_fit = std_y
            else:
                y_for_fit = std_y[:,i]
            lr = LinearRegression(fit_intercept=self.fit_intercept)
            lr.fit(hidden_outputs,y_for_fit)
            w = lr.coef_
            weights.append(w)
            if self.fit_intercept:
                b = lr.intercept_
                intercepts[i] = b
        weights = np.array(weights).T
        self.center = hidden_nodes_c
        self.width = hidden_nodes_w
        self.weight = weights
        self.intercepts = np.array(intercepts)

    def fit(self, x, y):
        self._fit(x, y)

    def _predict(self, x):
        std_x = self.standard(x, scale=self.scale_x )
        hidden_outputs = [
            np.exp(-np.sqrt(np.sum(np.power(std_x - self.center[i], 2), axis=1)) / (np.power(self.width[i], 2))) for
            i in range(self.hidden_layer_sizes)]
        hidden_outputs = np.array(hidden_outputs).T
        output = np.dot(hidden_outputs,self.weight)+self.intercepts
        pre = self.unstandard(output, self.scale_y)
        return pre

    def predict(self, x):
         pre = self._predict(x)
         if 1 in pre.shape:
             pre.shape = -1
         return pre


    def score(self, x, y):
        pred_y = self.predict(x)
        if len(y.shape) == 1:
            pred_y.shape = y.shape
        u = np.sum(np.power(pred_y-y,2))
        v = np.sum(np.power(pred_y-np.mean(pred_y, axis=0), 2))
        return 1-u/v

    def standard(self, data, scale=0):
        #数据标准化，返回标准化后的数据和数据的规格，该规格是2行的n列的矩阵列表，第一行表示减去的值，第二行表示长度
        if scale==0:
            a = np.min(data, axis=0)
            b = np.max(data, axis=0)-a
            a = a - self.times*b
            b = (2*self.times+1)*b
            std_data=(data-a)/b
            scale_data=[a,b]
            return std_data,scale_data
        else:
            std_data=(data-scale[0])/scale[1]
            return std_data

    def unstandard(self,std_data,scale_data):
        a=scale_data[0]
        b=scale_data[1]
        data = std_data*b+a
        return data



class RBFClassifier(RBFRegressor):
    def __init__(self, hidden_layer_sizes=10, times=0.3, fit_intercept=False):
        super().__init__(hidden_layer_sizes=hidden_layer_sizes, times=times, fit_intercept=fit_intercept)

    def fit(self, x, y):
        y = np.array(y).reshape(-1,1)
        self.ohe = OneHotEncoder(sparse=False)
        y_dist = self.ohe.fit_transform(y.reshape(-1,1))
        self._fit(x, y_dist)

    def predict_proba(self,x):
        raw_pro = self._predict(x)
        raw_pro = raw_pro - np.min(raw_pro,axis=1).reshape(-1,1)
        s = np.sum(raw_pro,axis=1)
        pro = raw_pro/s.reshape(-1,1)
        return pro

    def predict(self, x):
        pro = self._predict(x)        
        pre = self.ohe.inverse_transform(pro)
        return pre.reshape(-1)

    def score(self,x,y):
        pre_y = self.predict(x)
        n=len(pre_y)
        right = (pre_y==y).sum()
        return right/n







if __name__=='__main__':
    view = '分类回归分析'
    if view == '数值回归分析':
        # 数值回归分析
        s1 = RBFRegressor(hidden_layer_sizes=3,times=1,fit_intercept=True)
        s2 = MLPRegressor(hidden_layer_sizes=10, max_iter=800, activation='relu')
        # y=x1+2lnx2-3x3
        x = np.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12],[13,14,15],[16,17,18]])
        y = x[:, 0]+2*np.log(x[:, 1])-3*x[:, 2]
        y.shape = -1, 1
        test_x = np.array([[19,20,21],[22,23,24],[25,26,27]])
        test_y = test_x[:,0]+2*np.log(test_x[:,1])-3*test_x[:,2]
        test_y.shape = -1, 1
        s1.fit(x, y)
        s2.fit(x, y.reshape(-1))
        pre1_y = s1.predict(test_x)
        pre2_y = s2.predict(test_x)
        print('实际值：')
        print(test_y)
        print(' ')
        print('RBF网络输出和得分：')
        print(pre1_y)
        print(s1.score(test_x,test_y))
        print(' ')
        print('多层感知器输出和得分：')
        print(pre2_y)
        print(s2.score(test_x,test_y))
    elif view == '分类回归分析':
        # 分类回归分析
        s1 = RBFClassifier(hidden_layer_sizes=3,times=0.5)
        s2 = MLPClassifier(hidden_layer_sizes=10,max_iter=800,activation='relu')
        x = np.array([[10, 2, 3], [4, 15, 6], [-1.7, 8, 11], [-1.2, 9, 12], [5, 14, 5], [16, 7, 8]])
        y = np.array(['a','b','c','c','b','a'])
        y.shape = -1, 1
        #x第一列超过10就是a,2列超过是b，3列超过是c
        test_x = np.array([[1, 14, 2], [12, 3, 4], [5, 6, 17],[3,11,7]])
        test_y = np.array(['b','a','c','b'])
        s1.fit(x, y)
        s2.fit(x, y.reshape(-1))
        pre1_y = s1.predict(test_x)
        pre2_y = s2.predict(test_x)
        pro1 = s1.predict_proba(test_x)
        pro2 = s2.predict_proba(test_x)
        print('RBF网络输出概率和分类以及准确率：')
        print(pro1)
        print(pre1_y)
        print(s1.score(test_x, test_y))
        print('实际分类')
        print(test_y)
        print('多层感知器输出概率和分类以及准确率：')
        print(pro2)
        print(pre2_y)
        print(s2.score(test_x, test_y))



import numpy as np
import matplotlib.pyplot as plt
from .Methods import *
from .Scores import *

bar_color1 = 'darkturquoise'

def timing(aim = None):
    if aim is not None:
        now = dt.datetime.now()
        tmp = aim - now
        res = tmp.seconds
        print('waitting for {} seconds'.format(res))
        time.sleep(res)

def decompose_dim(arr, dim=0, *shape):
    """
    For input arrays, the specified dimensions are decomposed into the specified shapes
    
    Example:
    >>> arr = np.zeros([30, 200, 50])
    >>> d = decompose_dim(arr, 1, (20,-1))
    >>> d.shape
    (30, 20, 10, 50)
    """
    arr_shape = list(arr.shape)
    if np.iterable(shape[0]):
        shape = shape[0]
    if (-1 not in shape )&(np.prod(shape) != arr_shape[dim]):
        raise ValueError(f"The length of dimension {dim} must keep the same in decomposition")
    arr_shape[dim] = shape
    new_shape = []
    for i in arr_shape:
        if not np.iterable(i):
            new_shape.append(i)
        else:
            new_shape += list(i)
    return arr.reshape(new_shape)

def combine_dim(arr, *dim):
    """
    For the input array, reshape the specified dimension, making it a uniform dimension.
    
    Example:
    >>> arr = np.zeros([30, 10, 20, 50])
    >>> d = combine_dim(arr, 0, (0,1))
    >>> d.shape
    (300, 20, 50)
    """
    if np.iterable(dim[0]):
        dim = dim[0]
    else:
        dim = list(dim)
    shape = np.array(arr.shape)
    dim = np.sort(np.arange(len(shape))[dim])
    if ((dim[1:] - dim[:-1])!=1).any():
        warn_mes = "The input dimensions must be adjacent, otherwise other problems may result"
        np.warnings.warn(warn_mes, np.VisibleDeprecationWarning)
    shape[dim[-1]] = np.prod(shape[dim])
    shape = np.delete(shape, dim[:-1])
    return arr.reshape(shape)

def view(arr, **kwargs):
    """
    imshow an matrix
    """
    plt.colorbar(plt.imshow(arr, **kwargs)), plt.show()

def stat(arr, name='?'):
    """
    print the data loss rate, mean variance,
    minimum, mean, maximum
    """
    nan_num = np.isnan(arr).sum()
    nan_rate = nan_num/np.prod(arr.shape)
    vmin, vmean, vmax = np.nanmin(arr), np.nanmean(arr), np.nanmax(arr)
    sd = np.sqrt(np.nanmean((arr-vmean)**2))
    print(f"{name}: {nan_num*100:d}%, {sd}\n{vmin}, {vmean}, {vmax}")
    
if __name__ == "__main__":
    arr = np.zeros([300, 20, 50])
    d = decompose_dim(arr, 0, [-1, 5, 2])
    print(d.shape)
    d = combine_dim(d, [0,1])
    print(d.shape)
    



import numpy as np
import pandas as pd

def read_ascii(file,skip=1,dtype=float):
    f = open(file,'r')
    if skip>0:
        f.readlines(skip)
    data = []
    for line in f:
        d = [float(i) for i in line.split(' ') if i not in['','\n']]
        data.append(d)
    return np.array(data)


def maskout(x,y,points,groups=50):
    d = np.array(points)
    num = len(d)
    idx = np.argsort(d[:,0])
    d = d[idx,:]
    
    mask_shape = len(y),len(x)
    mask = np.zeros(mask_shape)
    size = np.round(num/groups)
    groups = np.int(np.ceil(num/size))
    x0 = np.zeros([groups])
    y0 = np.zeros([2,groups])

    for i in range(groups):
        st, ed = int(i*size), int(i*size+size)
        x0[i] = np.mean(d[st:ed,0])
        tmp = np.sort(d[st:ed,1])
        y0[0,i], y0[1,i] = tmp[0], tmp[-1]

    extend_x0 = np.hstack([np.array([x0[0]-0.001]),x0,np.array([x0[-1]+0.001])])
    p=np.argmin(np.abs(x.reshape(-1,1)-extend_x0),axis=1)
    start, end = np.where(p>0)[0][0],np.where(p<groups+1)[0][-1]

    for i in range(start,end+1):
        st = np.where(y>y0[0,p[i]-1])[0][0]
        ed = np.where(y<y0[1,p[i]-1])[0][-1]
        mask[st:ed+1,i] = 1

    mask = mask==0
    return mask

class Eof:
    
    def __init__(self,datasets,center=True):
        
        """
        data : an array with (n,***) shape. n is the time series grids
        return : an object with EOF analyzing methods
        """

        d = datasets.copy()
        n_times, *self._shape = d.shape
        d.shape = n_times, -1
        self._idx = np.isnan(d).any(axis=0)
        d = d[:,~self._idx]
        self.neofs = np.sum(~self._idx)
        if center:
            d = d-np.mean(d,0)
        cov = np.matmul(d.T,d)
        self._eig_val, self._eig_matrix = np.linalg.eig(cov)
        
        self._pcs = np.matmul(self._eig_matrix, d.T)

    def eofs(self,neofs=3):
        tmp = np.nan*np.zeros([neofs,np.prod(self._shape)])
        tmp[:,~self._idx] = self._eig_matrix[:,:neofs].T
        return tmp.reshape(neofs,*self._shape)

    def pcs(self,npcs=3):
        return self._pcs[:npcs].T

    def eigenvalues(self,neigs=None):
        if neigs is not None:
            r = self._eig_val[:neigs]
        else:
            r = self._eig_val
        return r

    def varianceFraction(self,neigs=None):
        return self.eigenvalues(neigs)/np.sum(self._eig_val)

class MultivariateEof:

    def __init__(self,datasets,center=True):
        
        data, info = self._merge_fields(datasets)
        self._shapes = info['shapes']
        self._slicers = info['slicers']
        self.solver = Eof(data,center=center)
        self.neofs = self.solver.neofs

    def eofs(self,neofs=3):
        modes = self.solver.eofs(neofs=neofs)
        return self._unwrap(modes)
    
    def pcs(self,npcs=3):
        return self.solver.pcs(npcs=npcs)

    def eigenvalues(self,neigs=None):
        return self.solver.eigenvalues(neigs)

    def varianceFraction(self,neigs=None):
        return self.solver.varianceFraction(neigs)


    def _unwrap(self, modes):
        """Split a returned mode field into component parts."""
        nmodes = modes.shape[0]
        modeset = [modes[:, slicer].reshape((nmodes,) + shape)
                   for slicer, shape in zip(self._slicers, self._shapes)]
        return modeset


    def _merge_fields(self, fields):
        """Merge multiple fields into one field.

        Flattens each field to (time, space) dimensionality and
        concatenates to form one field. Returns the merged array
        and a dictionary {'shapes': [], 'slicers': []} where the entry
        'shapes' is a list of the input array shapes minus the time
        dimension ans the entry 'slicers' is a list of `slice` objects
        that can be used to select each individual field from the merged
        array.

        """
        info = {'shapes': [], 'slicers': []}
        islice = 0
        for field in fields:
            info['shapes'].append(field.shape[1:])
            channels = np.prod(field.shape[1:])
            info['slicers'].append(slice(islice, islice + channels))
            islice += channels
        try:
            merged = np.concatenate(
                [field.reshape([field.shape[0], np.prod(field.shape[1:])])
                 for field in fields], axis=1)
        except ValueError:
            raise ValueError('all fields must have the same first dimension')
        return merged, info

def _get_shape(data,origin_scale,new_scale):
    origin_shape = data.shape
    origin_x_right = (origin_shape[1]-1)*origin_scale[0]
    origin_y_down = (origin_shape[0]-1)*origin_scale[1]
    s2, s1 = round(origin_x_right/new_scale[0])+1, round(origin_y_down/new_scale[1])+1
    s2 = s2+1 if s2*new_scale[0]<origin_x_right else s2
    s1 = s1+1 if s2*new_scale[1]<origin_y_down else s1
    return origin_shape,(s1,s2)

def inter(orgin,ref,n,shape):
    new_data = np.zeros(n)
    for i in range(n):
        index = int(i//ref)
        new_index = index if index == shape-1 else index+1
        delta = (i%ref)*(orgin[new_index]-orgin[new_index-1])/ref
        new_data[i] = orgin[index] + delta
    return new_data

def downscale(data,origin_scale,new_scale):
    origin_shape, new_shape = _get_shape(data,origin_scale,new_scale)
    new_data = np.zeros(new_shape)
    ref, conti = np.array(origin_scale)/np.array(new_scale), 0
    for j in range(0,new_shape[0]):
        index_origin = int(j//ref[1])
        if (conti==1)|(j%ref[1]==0):
            new_data[j, :], index_new = inter(data[index_origin, :], ref[0], new_shape[1], origin_shape[1]), j
            if index_origin+1<origin_shape[0]:
                gradi = inter(data[index_origin + 1, :], ref[0], new_shape[1], origin_shape[1]) - new_data[j, :]
        if (j%ref[1]+1)<=ref[1]:
            cul = new_data[index_new,:] if conti==0 else inter(data[index_origin, :], ref[0], new_shape[1], origin_shape[1])
            new_data[j,:] = cul + (j%ref[1])*gradi/ref[1]
            conti = 0
        elif (j%ref[1]+1)>ref[1]:
            conti = 1
            new_data[j, :] = new_data[index_new,:] + (j % ref[1]) * gradi/ref[1]
    return new_data

    


if __name__  == '__main__':
    l = np.zeros([80,82])
    print(downscale(l,[0.1,0.1],[0.05,0.05]).shape)



import sys
import os
import pickle as pk
import datetime as dt
import time

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import cartopy.crs as ccrs
import xarray as xr

from .Methods import *
from .Drawings import *
from .Scores import *
from . import usual


bar_color1 = 'darkturquoise'

def love():
    """
    A simple demo.
    """
    x = np.hstack([np.linspace(-1,-0.99,100),np.linspace(-0.99,0.99,800),np.linspace(0.99,1,100)])

    y = np.sqrt(1-x**2)+np.abs(x)
    plt.plot(x,y,color='orange',marker='.')

    y = -np.sqrt(1-x**2)+np.abs(x)
    plt.plot(x,y,'b.-')
    plt.legend([r'$y=\sqrt{1-x^2}+\left|x\right|$',r'$y=-\sqrt{1-x^2}+\left|x\right|$'])
    plt.show()




if __name__ == "__main__":
    arr = np.zeros([300, 20, 50])
    d = decompose_dim(arr, 0, [-1, 5, 2])
    print(d.shape)
    d = combine_dim(d, [0,1])
    print(d.shape)
    



import numpy as np


class ScoreFuns:


    def __init__(self, obs, pre, threshold=0.1, axis=None, eps=0.001):
        """
        No explain not yet.

        """
        if obs.shape != pre.shape:
            raise ValueError('The shape of "obs" and "pre" do not match')
        
        self.n_nums = np.sum(np.ones_like(obs),axis=axis)
        self.mask = np.isnan(obs)
        
        obs_ = np.where(obs >= threshold, 1., 0.)
        pre_ = np.where(pre >= threshold, 1., 0.)
        obs_[self.mask] = np.nan
        pre_[self.mask] = np.nan
        obs_t, pre_t = 1-obs_, 1-pre_
        
        self.hits = np.nansum(obs_*pre_,axis=axis)
        self.false_alarms = np.nansum(obs_t*pre_,axis=axis)
        self.misses =  np.nansum(obs_*pre_t,axis=axis)
        self.true_negative =  np.nansum(obs_t*pre_t,axis=axis)
        self.eps = eps

    def pod(self):
        """
        命中率：命中数/（实际总的降水数）
        """
        return self.hits/(self.hits+self.misses+self.eps)
        
    def far(self):
        """
        空报率：误警数/（总预报的降水数）
        """
        return self.false_alarms/(self.hits+self.false_alarms+self.eps)
        
        
    def hss(self):
        num=2*(self.hits*self.true_negative - self.misses*self.false_alarms)
        den = (self.misses**2 + self.false_alarms**2 + 2*self.hits*self.true_negative
                   + (self.misses + self.false_alarms)*(self.hits + self.true_negative))
        return num/(den+self.eps)

    def acc(self):
    	
        return (self.hits+self.true_negative)/self.n_nums

    def ets(self):
        num = (self.hits + self.false_alarms) * (self.hits + self.misses)
        den = self.hits + self.misses + self.false_alarms + self.true_negative
        Dr = num / den
        return (self.hits - Dr) / (self.hits + self.misses + self.false_alarms - Dr+self.eps)

    def bias(self, vmax=2):
        """
        偏差评分(Bias score)主要用来衡量模式对某一量级降水的预报偏差, 该评分在数值上等于预报区域内满足某降水阈值的总格点数与对应实况降水总格点数的比值。
        是用来反映降水总体预报效果的检验方法。
        当BIAS>1时, 表示预报结果较实况而言偏湿;
        当BIAS < 1时, 表示预报结果较实况而言偏干;
        当BIAS=1时, 则表示预报偏差为0, 即预报技巧最高。

        由于BIAS评分主要是用于衡量预报区域内满足某降水阈值的预报技巧, 并不能衡量降水的准确率, 因此还需引入公平技巧评分(Equitable Threat Score, ETS)用于衡量对流尺度集合预报的预报效果。
        ETS评分表示在预报区域内满足某降水阈值的降水预报结果相对于满足同样降水阈值的随机预报的预报技巧, 因此该评分有效地去除了随机降水概率对评分的影响, 相对而言更加公平、客观。
        根据定义, 如果ETS>0时, 表示对于某量级降水来说预报有技巧; 如果ETS≤0时则表示预报没有技巧; 如果ETS=1则表示该预报为完美预报。
        """
        bia = (self.hits + self.false_alarms)/(self.hits + self.misses + 0.000001)
        bia[bia>vmax] = vmax
        return bia
        
    def ts(self):
        return self.hits/(self.hits + self.false_alarms+self.misses+self.eps)


    def get(self,*args):
        if isinstance(args[0],(tuple,list)):
            args = args[0]
        dic = dict(ts=self.ts, pod=self.pod, far=self.far,
                   hss=self.hss, acc=self.acc, ets=self.ets, bias=self.bias)
        return [dic.get(s.lower(),lambda x: None)() for s in args]


        

if __name__=='__main__':
    pass





























import numpy as np
from random import randint
from sklearn.neural_network import MLPRegressor, MLPClassifier
from sklearn.cluster import KMeans
from time import sleep


class GA_RBFRegressor:
    def __init__(self, pop=5, max_iter=200, hidden_layer_size=10,  times=0.5, fit_intercept=False):
        self.hidden_layer_size = hidden_layer_size
        self.times = times
        self.pop = pop
        self.max_iter = max_iter
        self.fit_intercept = fit_intercept

    def _fit(self, x, y):
        num = len(x)
        if num>1000:
            rest_time = 10
        else:
            rest_time = 0
        print(rest_time)
        std_x, self.scale_x = self.standard(x)
        std_y, self.scale_y = self.standard(y)
        if len(y.shape) == 1:
            self.output_layer_size = 1
        else:
            self.output_layer_size = y.shape[1]
        # 确定隐层的核中心和基宽
        km = KMeans(n_clusters=self.hidden_layer_size,n_jobs=3)
        km.fit(std_x)
        self.center = [i for i in km.cluster_centers_]
        self.width = [np.mean(np.sqrt(np.sum(np.power(std_x-i, 2), axis=1))) for i in km.cluster_centers_]
        # 用遗传算法确定隐层到输出层的权重和截距（偏置）
        # 初始化种群
        population_w, population_b = list(), list()
        for j in range(self.pop):
            one_w = np.random.uniform(-1, 1, [self.hidden_layer_size, self.output_layer_size])
            population_w.append(one_w)
            one_b = np.random.uniform(-1, 1, self.output_layer_size)
            population_b.append(one_b)
        # 开始遗传优化
        best_pop = [np.inf,0,0]
        iteration = 0
        while iteration <= self.max_iter:
            new_pop_w, new_pop_b=list(),list()
            # 计算适应度并记录最好个体
            pop_fitness = [self.fitness(std_x, std_y, population_w[i], population_b[i]) for i in range(len(population_w))]
            if min(pop_fitness) < best_pop[0]:
                n = pop_fitness.index(min(pop_fitness))
                best_pop = [pop_fitness[n], population_w[n], population_b[n]]
            for i in range(int(self.pop/2)):
                # 选择操作
                x, y = self.select(pop_fitness)
                # 将个体建设为染色体
                person1_w, person2_w = population_w[x].reshape(-1), population_w[y].reshape(-1)
                # 交叉互换
                person1_w, person2_w = self.cross(person1_w, person2_w) 
                # 变异
                person1_w = self.mut(person1_w)
                person2_w = self.mut(person2_w)
                # 将染色体进化为个体
                person1_w.shape = self.hidden_layer_size, self.output_layer_size
                person2_w.shape = self.hidden_layer_size, self.output_layer_size
                # 构建新的种群
                new_pop_w += [person1_w,person2_w]

                if self.fit_intercept:
                    person1_b, person2_b = self.matrix_to_vector(population_b[x]), self.matrix_to_vector(population_b[y])
                    person1_b, person2_b = self.cross(person1_b, person2_b)
                    person1_b = self.mut(person1_b)
                    person2_b = self.mut(person2_b)
                    person1_b = np.array(person1_b)
                    person2_b = np.array(person2_b)
                    new_pop_b += [person1_b,person2_b]
                else:
                    new_pop_b += [0, 0]
                
            new_pop_w.append(best_pop[1])
            new_pop_b.append(best_pop[2])
            population_w,population_b=new_pop_w,new_pop_b
            iteration += 1
            if iteration%20==0 and rest_time!=0:
                print('({}/{})resting'.format(iteration,self.max_iter))
                sleep(rest_time)     
        else:
            # 遗传进化完成，计算适应度并记录最好个体
            pop_fitness = [self.fitness(std_x, std_y, population_w[i], population_b[i]) for i in
                           range(len(population_w))]
            if min(pop_fitness) < best_pop[0]:
                n = pop_fitness.index(min(pop_fitness))
                best_pop = [pop_fitness[n], population_w[n], population_b[n]]
        self.weight, self.intercepts = best_pop[1], best_pop[2]


    def fit(self, x, y):
        self._fit(x, y)

    def _predict(self, x):
        std_x = self.standard(x, scale=self.scale_x )
        hidden_outputs = [
            np.exp(-np.sqrt(np.sum(np.power(std_x - self.center[i], 2), axis=1)) / (np.power(self.width[i], 2))) for
            i in range(self.hidden_layer_size)]
        hidden_outputs = np.array(hidden_outputs).T
        if not self.fit_intercept:
            self.intercepts = 0
        output = np.dot(hidden_outputs,self.weight)+self.intercepts
        pre = self.unstandard(output, self.scale_y)
        return pre

    def predict(self, x):
         pre = self._predict(x)
         if 1 in pre.shape:
             pre.shape = -1
         return pre


    def score(self, x, y):
        pred_y = self.predict(x)
        if len(y.shape) == 1:
            pred_y.shape = y.shape
        u = np.sum(np.power(pred_y-y,2))
        v = np.sum(np.power(pred_y-np.mean(pred_y, axis=0), 2))
        return 1-u/v

    def standard(self, data, scale=0):
        #数据标准化，返回标准化后的数据和数据的规格，该规格是2行的n列的矩阵列表，第一行表示减去的值，第二行表示长度
        if scale==0:
            a = np.min(data, axis=0)
            b = np.max(data, axis=0)-a
            a = a - self.times*b
            b = (2*self.times+1)*b
            std_data=(data-a)/b
            scale_data=[a,b]
            return std_data,scale_data
        else:
            std_data=(data-scale[0])/scale[1]
            return std_data

    def unstandard(self,std_data,scale_data):
        a=scale_data[0]
        b=scale_data[1]
        data = std_data*b+a
        return data

    # 遗传算法的一些操作所需函数
    def fitness(self,x,y,w,b):
        hidden_outputs = [
            np.exp(-np.sqrt(np.sum(np.power(x - self.center[i], 2), axis=1)) / (np.power(self.width[i], 2))) for
            i in range(self.hidden_layer_size)]
        hidden_outputs = np.array(hidden_outputs).T
        if not self.fit_intercept:
            b = 0
        pre = np.dot(hidden_outputs, w)+b
        if len(y.shape)==1:
            pre.shape = y.shape
        dist =np.sqrt(np.sum(np.power(pre-y,2)))
        return dist
    
    def select(self,f,d=15):
        # 根据适应度选择两个个体，d为选择程度，适应度越小越容易被选中
        n=len(f)
        f=np.array(f)
        p=pow(1/f,d)
        p=p/sum(p)
        # 以上求倒数乘方并归一化得出选取概率，以下产生随机数选取
        ra=np.random.random()
        s=0  # 累积概率
        for i in range(n):
            s+=p[i]
            if ra<=s:
                x=i
                break
        p[x]=0
        p=p/sum(p)
        ra=np.random.random()
        s=0
        for i in range(n):
            s+=p[i]
            if ra<=s:
                y=i
                break
        return x,y

    def cross(self, x, y):
        n = len(x)
        a, b = randint(0,n-1), randint(0,n-1)
        [a,b] = np.sort([a,b])
        (x[a:b+1], y[a:b+1]) = (y[a:b+1], x[a:b+1])
        return x, y

    def mut(self,x):
        n=len(x)
        a,b=randint(0,n-1),randint(0,n-1)
        [a,b]=np.sort([a,b])
        random_value = np.random.uniform(-1,1,[1,b-a+1])
        x[a:b+1] = np.array(x[a:b+1]) + random_value
        return x






class GA_RBFClassifier(GA_RBFRegressor):
    def __init__(self,pop=5, max_iter=200, hidden_layer_size=10, times=0.5, fit_intercept=False):
        super().__init__(pop=pop, max_iter=max_iter, hidden_layer_size=hidden_layer_size, times=times, fit_intercept=fit_intercept)

    def fit(self, x, y):
        cate = list()
        num_cate = 0
        for i in y:
            if i not in cate:
                num_cate += 1
                cate.append(i)
        y_dist = -1000*np.ones([len(y), num_cate])
        for i in range(len(y)):
            y_dist[i, cate.index(y[i])] = 1000
        self.classes_ = np.array(cate)
        self._fit(x, y_dist)

    def predict_proba(self,x):
        raw_pro = self._predict(x)
        for i in range(len(raw_pro)):
            raw_pro[i, :] = raw_pro[i, :]-np.min(raw_pro[i, :])
            s = np.sum(raw_pro[i, :])
            raw_pro[i, :] = raw_pro[i, :]/s
        return raw_pro

    def predict(self, x):
        pro = self._predict(x)
        most_pro_index = np.argmax(pro, axis=1)
        cate = list()
        for i in most_pro_index:
            cate.append(self.classes_[i])
        pre = np.array(cate)
        if 1 in pre.shape:
            pre.shape=-1
        return pre

    def score(self,x,y):
        pre_y = self.predict(x)
        right=0
        n=len(pre_y)
        for i in range(n):
            if pre_y[i] == y[i]:
                right += 1
        return right/n



if __name__=='__main__':
    view = '分类回归分析'
    if view == '数值回归分析':
        # 数值回归分析
        s1 = GA_RBFRegressor(hidden_layer_size=3, pop=8, max_iter=1000, times=0.6, fit_intercept=True)
        s2 = MLPRegressor(hidden_layer_sizes=10, max_iter=800, activation='relu')
        # y=x1+2lnx2-3x3
        x = np.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12],[13,14,15],[16,17,18]])
        y = x[:, 0]+2*np.log(x[:, 1])-3*x[:, 2]
        y.shape = -1, 1
        test_x = np.array([[19,20,21],[22,23,24],[25,26,27]])
        test_y = test_x[:,0]+2*np.log(test_x[:,1])-3*test_x[:,2]
        test_y.shape = -1, 1
        s1.fit(x, y)
        s2.fit(x, y)
        pre1_y = s1.predict(test_x)
        pre2_y = s2.predict(test_x)
        print('实际值：')
        print(test_y)
        print(' ')
        print('RBF网络输出和得分：')
        print(pre1_y)
        print(s1.score(test_x,test_y))
        print(' ')
        print('多层感知器输出和得分：')
        print(pre2_y)
        print(s2.score(test_x,test_y))
    elif view == '分类回归分析':
        # 分类回归分析
        s1 = GA_RBFClassifier(hidden_layer_size=3,pop=8, max_iter=500, times=0.6, fit_intercept=False)
        s2 = MLPClassifier(hidden_layer_sizes=[10,2],max_iter=800,activation='relu')
        x = np.array([[10, 2, 3], [4, 15, 6], [7, 8, 11], [1, 9, 12], [5, 14, 5], [16, 7, 8]])
        y = np.array(['a','b','c','c','b','a'])
        y.shape = -1, 1
        #x第一列超过10就是a,2列超过是b，3列超过是c
        test_x = np.array([[1, 14, 2], [12, 3, 4], [5, 6, 17],[3,11,7]])
        test_y = np.array(['b','a','c','b'])
        test_y.shape = -1, 1
        s1.fit(x, y)
        s2.fit(x, y)
        pre1_y = s1.predict(test_x)
        pre2_y = s2.predict(test_x)
        pro1 = s1.predict_proba(test_x)
        pro2 = s2.predict_proba(test_x)
        print('RBF网络输出概率和分类以及准确率：')
        print(pro1)
        print(pre1_y)
        print(s1.score(test_x, test_y))
        print('实际分类')
        print(test_y)
        print('多层感知器输出概率和分类以及准确率：')
        print(pro2)
        print(pre2_y)
        print(s2.score(test_x, test_y))



import matplotlib.pyplot as plt
import numpy as np
import datetime as dt
import os
import warnings

import cartopy.crs as ccrs
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter
import cartopy.feature as cfeature
import matplotlib.ticker as mticker
import matplotlib.patches as patches
import xarray as xr


def _get_unique(dic1, dic2):
    d = {}
    for i in dic1.keys():
        if i not in dic2.keys():
            d.update({i: dic1[i]})
    return d


class map:

    def __init__(self, x=[70,140], y=[0,60], xticks=None, yticks=None, center=110, style='minor', **kwargs):
        """Create a geographical map object.

        The map object can be initializated one time, and be drew many times. Method
        calls are used to set paramters or drawing. Some paramters are shared by all
        the methods of this map object.

        **Arguments:**

        *x*
            A list that contains three or two items. The first two iterms are the
            beginning and the ending of the longitudes. The third iterm is optional
            which means the step from the beginning to the end, if it is not given, we
            would set it 1 as default.

        *y*
            The same as x, but was lattitude's.

        **Optional arguments:**

        *xticks*
            A iterable object. If labeled, label the specific longitudes.If is default,
            draw some label lines in the map.

        *yticks*
            The same as xticks, but was lattitude's.

        *center*
            If the projection is the default PlateCarree, center is the center longitude of this projection.

        **Returns:**
            A basemap object.

        *Other parameters:*
        
        *subplot*
            subplot infomation, set to (1,1,1) as default.

        *projection*
            A cartopy.crs object. It is the projection method of map.

        *transform*
            A cartopy.crs object that transform the data projection,
            only used when filling the data.

        *crs*
            A cartopy.crs, was used to define a map.

        *coast*
            Bool. Whether to draw the coast.(True as default)
            
        *coast_linewidth*   : default 1
        *coast_color*   : default "black"
        *coast_linestyle*  : default '-'
        *coast_resolution*  : default "110m"

        *borders*
            Bool. Whether to draw the border of the countries.(False as default)
            
        *border_linewidth*  : 1
        *border_linestyle*  : '-'
        *border_color*  : 'k'
        *border_resolution*  : "110m"

        *land*
            Bool. Whether to draw the border of the countries.(False as default)
            
        *land_linewidth*  : 1
        *land_linestyle*  : '-'
        *land_color*  : 'k'
        *land_resolution*  : "110m"
        
        *rivers*
            Bool. Whether to load the rivers.(False as default)     
        *river_linewidth*  : 1
        *river_linestyle*  : '-'
        *river_color*  : 'blue'
        *river_resolution*  : '110m',

        *lake*:  the same as *rivers*

        *lonlat*
            Bool. Whether to label the longitude and latitude.(True as default)

        *grid*: the same as borders

        *labels*
            A list that defines where to draw the lonlat labels.
            Set ['left','bottom'] as default.



        **Examples**
            >>> m = map([70,130,2],[15,55,2])
            >>> m.load_china()
            >>> d = np.arange(31*21).reshape(21,31)
            >>> c = m.contourf(d,cmap=plt.cm.bwr)
            >>> m.show()
            
            Tips:
            If we need to draw some other maps, we can run as:
            >>> m.subplot(2,3,1)
            >>> m.load_province()
            >>> m.contourf(d)
            Or:
            >>> m.axes([0.1,0.1,0.5,0.5])
            >>> m.contourf(d)
            
            If we need to add a shapefile, it runs as:
            >>> m.add_shape("./shapefiles/TPBoundary_2500.shp")
            
            If we need other projection, it should be define in the map initialization:
            >>> import cartopy.crs as ccrs
            >>> m = map([70,130,2],[15,55,2],projection=ccrs.EquidistantConic(110,60))
            
        For more usage, use help(func) or contract me.
        My QQ number is 1439731362.

        """
        if np.max(np.abs(y))>90:
            ErrorMess = "latitude must be in [-90,90]"
            raise ValueError(ErrorMess)
        self._para = {'projection': ccrs.PlateCarree(center), 'transform': ccrs.PlateCarree(),
                      'crs': ccrs.PlateCarree(),
                      'coast': True, 'coast_linewidth': 1, 'coast_color': 'black', 'coast_linestyle': '-',
                      'coast_resolution': '110m',
                      'borders': False, 'border_linewidth': 1, 'border_linestyle': '-', 'border_color': 'k',
                      'border_resolution': '110m',
                      'land': False, 'land_linewidth': 1, 'land_linestyle': '-', 'land_color': 'gray',
                      'land_resolution': '110m',
                      'rivers': False, 'river_linewidth': 1, 'river_linestyle': '-', 'river_color': 'blue',
                      'river_resolution': '110m',
                      'lakes': False, 'lake_linewidth': 0.8, 'lake_linestyle': '-', 'lake_color': 'blue',
                      'lake_resolution': '110m',
                      'lonlat': True, 'grid': False, 'grid_linewidth': 0.9, 'grid_linestyle': ':', 'grid_color': 'k',
                      'labels': ['left', 'bottom'],

                      }
        self._para.update(kwargs)
        x, y = list(x), list(y)
        x[1], y[1] = x[1] + 0.0001, y[1] + 0.0001
        if x[1] > 360:
            x[1] = 359.99
        self.xlims, self.ylims = x[:2], y[:2]
        X, Y = np.arange(*x), np.arange(*y)
        
        self.x, self.y = X, Y
        self.xticks = np.round(np.linspace(*self.xlims, 5)) if xticks is None else np.array(xticks)
        self.xticks[self.xticks > 180] = self.xticks[self.xticks > 180] - 360
        self.xticks = np.sort(self.xticks)
        self.yticks = np.round(np.linspace(*self.ylims, 5)) if yticks is None else np.array(yticks)
        self.MainWidget = False
        self.style = 'minor'

        
    def set_style(self, xnums=5, ynums=5, **kwargs):
        from matplotlib.ticker import AutoMinorLocator
        kw = dict(which='major', width=1.1, length=5)
        kw.update(kwargs)
        
        def __set_style(xnums=xnums, ynums=ynums, **kw):
            kw = dict(which='major', width=1.1, length=5)
            self.ax.tick_params(**kw)
            self.ax.xaxis.set_minor_locator(AutoMinorLocator(xnums))
            self.ax.yaxis.set_minor_locator(AutoMinorLocator(ynums))

        self.__style_func = __set_style

    def __repr__(self):
        a, b = round(self.xlims[0], 2), round(self.xlims[1], 2)
        if b == 359.99:
            b = 360
        c, d = round(self.ylims[0], 2), round(self.ylims[1], 2)
        string = "<oyl.map>\nxlims :   {:} - {:}\nylims :   {:} - {:}".format(a, b, c, d)
        return string

    def axes(self, *args, projection=None):
        """
        See matplotlib.pyplot.axes
        """
        proj = projection if projection else self._para['projection']
        self.ax = plt.axes(*args, projection=proj)
        self.MainWidget = False
        return self.ax

    def subaxes(self, *args, projection=None):
        """
        See matplotlib.pyplot.axes
        """
        proj = projection if projection else self._para['projection']
        self.subax = plt.axes(args, projection=proj)
        return self.subax

    def small_map(self, loc=[0.795, 0.005, 0.2, 0.3], extent=[105,125,0,25], shape=True, custom=False, projection=None):
        """
        Draw a small map in the main map.
        
        *loc*
            The relative location of the whole main map.

        *extent*
            The extent of the small map

        *shape*
            Whether to add the China shapefile

        return:
            A GeoAxes
        """
        if not self.MainWidget:
            raise ValueError("A complete main map must be defined")

        a, b, c, d = loc
        x, y, w, h = self.ax.get_position().bounds

        X, Y = x + w*a, y + h*b
        W, H = c*w, d*h
        ax = self.subaxes(X, Y, W, H)
        ax.set_extent(extent, ccrs.PlateCarree())
        if not custom:
            self.load_china(which='sub')
        return ax

    def extent(self, extents=None, crs = None):
        crs = crs if crs else self._para['crs']  
        if not extents:
            extents = self.xlims[0], self.xlims[1], self.ylims[0], self.ylims[1]
        self.ax.set_extent(extents, crs=crs)

    def figure(self, *args, **kwargs):
        """
        See matplotlib.pyplot.figure
        """
        self.__need_subplot, self.__extend = True, True
        self.fig = plt.figure(*args, **kwargs)
        return self.fig

    def subplot(self, *args, **kwargs):
        """
        See matplotlib.pyplot.subplot
        """
        self.MainWidget = False
        self.__subplot(*args, **kwargs)

    def __subplot(self, *args, **kwargs):

        if not self.MainWidget:
            
            self.fig = plt.gcf()
            if 'projection' not in kwargs:
                kwargs.update({"projection":self._para['projection']})
            self.ax = self.fig.add_subplot(*args, **kwargs)
            if self._para['coast']:
                self.ax.coastlines(resolution=self._para['coast_resolution'], lw=float(self._para['coast_linewidth']),
                                   color=self._para['coast_color'], linestyle=self._para['coast_linestyle'])
            if self._para['borders']:
                self.ax.add_feature(cfeature.BORDERS.with_scale(self._para['border_resolution']),
                                    linestyle=self._para['border_linestyle'],
                                    lw=float(self._para['border_linewidth']), edgecolor=self._para['border_color'])
            if self._para['land']:
                self.ax.add_feature(cfeature.LAND.with_scale(self._para['land_resolution']),
                                    linestyle=self._para['land_linestyle'],
                                    lw=float(self._para['land_linewidth']), color=self._para['land_color'])
            if self._para['rivers']:
                self.ax.add_feature(cfeature.RIVERS.with_scale(self._para['river_resolution']),
                                    linestyle=self._para['river_linestyle'],
                                    lw=float(self._para['river_linewidth']), color=self._para['river_color'])
            if self._para['lakes']:
                self.ax.add_feature(cfeature.LAKES.with_scale(self._para['lake_resolution']),
                                    linestyle=self._para['lake_linestyle'],
                                    lw=float(self._para['lake_linewidth']), color=self._para['lake_color'])

            if self._para['lonlat']:
                draw_labels = False if isinstance(self._para["projection"], ccrs.PlateCarree) else True
                if not self._para['grid']:
                    self._para['grid_linewidth'] = 0
                gl = self.ax.gridlines(draw_labels=draw_labels, linewidth=self._para['grid_linewidth'],
                                       linestyle=self._para['grid_linestyle'], color=self._para['grid_color'],
                                       alpha=0.8,
                                       xlocs=self.xticks, ylocs=self.yticks)
                for la in ['left', 'bottom', 'top', 'right']:
                    if la not in self._para["labels"]:
                        exec(f"gl.{la}_labels=False")
                if isinstance(self._para["projection"], ccrs.PlateCarree):
                    gl.xlocator = mticker.FixedLocator(self.xticks)
                    gl.ylocator = mticker.FixedLocator(self.yticks)

                    self.ax.set_xticks(self.xticks, crs=self._para['crs'])
                    self.ax.set_yticks(self.yticks, crs=self._para['crs'])
                    self.ax.xaxis.set_major_formatter(LongitudeFormatter(zero_direction_label=True))
                    self.ax.yaxis.set_major_formatter(LatitudeFormatter())
            self.extent()
            if self.style=='minor':
                try:
                    self.__style_func()
                except:
                    self.set_style()
                    self.__style_func()
            self.MainWidget = True

    def __contourf(self, *args, **kwargs):
        if len(args) == 1:
            args = self.x, self.y, args[0]

        dic = {'cbar': True, 'fmt': '%d', 'pad': 0.08, 'fraction': 0.04,
               'location': 'right', 'cax': None}
        kw = _get_unique(kwargs, dic)
        dic.update(kwargs)
        if 'transform' not in kw:
            kw.update({"transform":self._para['transform']})

        self.contourf_map = self.ax.contourf(*args, **kw)
        if dic['cbar']:
            ori = {'bottom': 'horizontal', 'right': 'vertical'}
            self.colorbar = plt.colorbar(self.contourf_map, orientation=ori[dic['location']], format=dic['fmt'],
                                         pad=dic['pad'], fraction=dic['fraction'], cax=dic['cax'])
        return self.contourf_map

    def contourf(self, *args, **kwargs):
        """
        This function will automatically add latitude and longitude if they are not given.
        This function don' need to consider the transform as it is automatically defined.
        The colorbar will be automatically added.
        The colorbar related parameters are:
        *cbar*
            Bool. Whether to add the colorbar.(True as default)
        *location*
            "bottom" or "right", the location of the cbar.  
        More arguments can be seen for matplotlib.pyplot.contourf and matplotlib.pyplot.colorbar
        """
        if not self.MainWidget:
            self.__subplot(111)
        return self.__contourf(*args, **kwargs)

    def __contour(self, *args, **kwargs):
        if len(args) == 1:
            args = self.x, self.y, args[0]

        dic = {'clabel': True, 'fmt': '%d', 'inline': True, 'linewidths': 0.85,
               'fontsize': 8, 'color': 'k'}
        kw = _get_unique(kwargs, dic)
        if ('cmap' in kw.keys()) & ('colors' in kw.keys()):
            kw.pop('colors')
        dic.update(kwargs)
        if 'transform' not in kw:
            kw.update({"transform":self._para['transform']})
        self.contour_map = self.ax.contour(*args, linewidths=dic['linewidths'], **kw)

        if dic['clabel']:
            cb = plt.clabel(self.contour_map, inline=dic['inline'], fontsize=dic['fontsize'],
                            colors=kwargs.get('color', kwargs.get('colors', None)), fmt=dic['fmt'])
        return self.contour_map

    def contour(self, *args, **kwargs):
        """
        This function will automatically add latitude and longitude if they are not given.
        This function don' need to consider the transform as it is automatically defined.
        The labels will be automatically added.
        The labels related parameters are:
        *clabel*
            Bool. Whether to add the labels.(True as default)
        More arguments can be seen for matplotlib.pyplot.contour and matplotlib.pyplot.clabel
        """
        if not self.MainWidget:
            self.__subplot(111)
        return self.__contour(*args, **kwargs)

    def ttest_plot(self, logic_matrix, x=None, y=None, **kwargs):
        """
        This function is uesd for drawing points for the area who passing inspection.
        Not recommended. This function will be abandon in the futrue.
        Use other funcions such as scatter or plot instead.
        """

        dic = {'color': 'k', 'marker': 'o', 'linestyle': '',
               'alpha': 0.8, 'markersize': 1, }

        X, Y = self.x if x is None else x, self.y if y is None else y
        xx, yy = np.meshgrid(X, Y)
        x, y = xx[logic_matrix], yy[logic_matrix]
        self.ax.plot(x, y, color=dic['color'], marker=dic['marker'], linestyle=dic['linestyle'],
                     alpha=dic['alpha'], markersize=dic['markersize'], transform=self._para['transform'])
        return

    def __pcolor(self, *args, **kwargs):
        if len(args) == 1:
            args = self.x, self.y, args[0]


        dic = {'cbar': True, 'fmt': '%d', 'pad': 0.08, 'fraction': 0.04,
               'location': 'right', 'cax': None, 'extend': 'neither'}
        kw = _get_unique(kwargs, dic)
        dic.update(kwargs)
        if 'transform' not in kw:
            kw.update({"transform":self._para['transform']})

        self.pcolor_map = self.ax.pcolor(*args, **kw)
        if dic['cbar']:
            ori = {'bottom': 'horizontal', 'right': 'vertical'}
            cb = plt.colorbar(self.pcolor_map, orientation=ori[dic['location']], format=dic['fmt'],
                              pad=dic['pad'], fraction=dic['fraction'], cax=dic['cax'], extend=dic['extend'])
        return self.pcolor_map

    def pcolor(self, *args, **kwargs):
        """
        This function will automatically add latitude and longitude if they are not given.
        This function don' need to consider the transform as it is automatically defined.
        The colorbar will be automatically added.
        The colorbar related parameters are:
        *cbar*
            Bool. Whether to add the colorbar.(True as default)
        *location*
            "bottom" or "right", the location of the cbar.  
        More arguments can be seen for matplotlib.pyplot.pcolor and matplotlib.pyplot.colorbar
        """
        if not self.MainWidget:
            self.__subplot(111)
        return self.__pcolor(*args, **kwargs)

    def __quiver(self, *args, skip=None, legend=False, **kwargs):

        if len(args) == 2:
            args = [self.x, self.y, args[0], args[1]]

        dic = dict(X=1.05,Y=0.45,U=10,label='10m/s',angle=90,labelpos='N')
        kw = _get_unique(kwargs, dic)
        dic.update(kwargs)
        for key in kw:
            dic.pop(key)
        if 'transform' not in kw:
            kw.update({"transform":self._para['transform']})

        
        if skip is not None:
            xs, ys = skip
            args[0] = args[0][::xs]
            args[1] = args[1][::ys]
            args[2] = args[2][::ys, ::xs]
            args[3] = args[3][::ys, ::xs]
        self.quiver_map = self.ax.quiver(*args, **kw)
        if legend:
            plt.quiverkey(self.quiver_map, **dic)

    def quiver(self, *args, skip=None, legend=False, **kwargs):
        """
        This function will automatically add latitude and longitude if they are not given.
        This function don' need to consider the transform as it is automatically defined.
        The labels will be automatically added.
        The other parameters are:
        *skip*
            A list that controls the density of x and y directions.
            If the resolution of the data is so high that the wind field is too dense to see clearly, we need to sparse it.
        *legend*
            Bool. Whether to add the wind scale.(False as default)

        **Examples**
        >>>m.quiver(u,v,skip=[3,3],scale=200,legend=True,X=1.04,Y=0.45,U=4,label='4m/s',angle=90,labelpos='N')
        
        More arguments can be seen for matplotlib.pyplot.quiver and matplotlib.pyplot.quiverkey
        """
        if not self.MainWidget:
            self.__subplot(111)
        return self.__quiver(*args, skip=skip, legend=legend, **kwargs)

    def __streamplot(self, *args, skip=None, **kwargs):
        if len(args) == 2:
            args = [self.x, self.y, args[0], args[1]]
        kw = dict(color='k', linewidth=0.6, arrowsize=0.8, arrowstyle='<-', transform=self._para['transform'])
        kw.update(kwargs)
        
        
        if skip is not None:
            xs, ys = skip
            args[0] = args[0][::xs]
            args[1] = args[1][::ys]
            args[2] = args[2][::ys, ::xs]
            args[3] = args[3][::ys, ::xs]
        self.stream_map = self.ax.quiver(*args, **kw)

    def streamplot(self, *args, skip=None, **kwargs):
        """
        See matplotlib.pyplot.streamplot
        """
        if not self.MainWidget:
            self.__subplot(111)
        return self.__streamplot(*args, skip=skip, **kwargs)

    def plot(self, *args, **kwargs):
        """
        See matplotlib.pyplot.plot
        """
        if not self.MainWidget:
            self.__subplot(111)
        if 'transform' not in kwargs:
            kwargs.update({"transform":self._para['transform']})
        return self.ax.plot(*args, **kwargs)

    def add_shape(self, filename, which='main', linewidth=1, linestyle='-', encoding='gbk', rec=-1, method=None, **kwargs):
        """
        There are two ways to add a shapefile to the map.
        One is to use cartopy.io to read the shp file and cartopy.feature to add it.
        The other is to use pyshp to read the shp file and add a matplotlib.patches.
        Some shp files are difficult to read. This function would try another method after one fails.

        **Arguments:**

        *filename*
            The shp file path. The three files (.shp, .shx, .dbf) are needed.

        **Optional arguments:**

        *encoding*
            The shp file encoding, only useful when using the pyshp's Reader.

        *method*
            Which way to load the shp. (None as default as the cartopy, and set to the other if it fails.)
            It can be any object that can make logical judgment positive to use pyshp to read the shp file and add a matplotlib.patches.

        *rec*
            A number. Sometimes the shp file contains many records which represents different areas.
            Each area has its ID number. This arguments tells the map which area to draw.
            It is -1 by default as all the areas are shown. (Only useful for the pyshp method)
               
        More arguments can be seen for cartopy.feature.ShapelyFeature, matplotlib.patches.Polygon
        """
        if not self.MainWidget:
            self.__subplot(111)
        ax = self.ax if which=='main' else self.subax
        from cartopy.io.shapereader import Reader
        dic = dict(edgecolor='k', facecolor=None)
        dic.update(kwargs)
        if not method:
            try:
                dic0 = dic.copy()
                reader = Reader(filename)
                dic0['facecolor'] = dic['facecolor'] if dic['facecolor'] else 'none'
                enshicity = cfeature.ShapelyFeature(reader.geometries(), self._para['transform'], **dic0)
                ax.add_feature(enshicity, linewidth=linewidth, linestyle=linestyle)
                reader.close()
            except:
                method = 1
        if method:
            import matplotlib.patches as patches
            from shapefile import Reader
            if isinstance(rec, int) and (rec > 0):
                rec = [rec]
            sf = Reader(filename, encoding=encoding)
            fill = True if dic['facecolor'] else False
            for i, sh in enumerate(sf.shapes()):
                if (rec == -1) or (i in rec):
                    points = np.array(sh.points)
                    if len(sh.parts) == 2:
                        ax.add_patch(
                            patches.Polygon(points[sh.parts[1]:], linestyle=None, transform=self._para['transform'],
                                            color=dic.get('color', dic['edgecolor']), fill=fill,
                                            facecolor=dic['facecolor']))
                        points = points[:sh.parts[1]]
                    ax.add_patch(patches.Polygon(points, linewidth=linewidth, linestyle=linestyle,
                                                      transform=self._para['transform'],
                                                      color=dic.get('color', dic['edgecolor']), fill=fill,
                                                      facecolor=dic['facecolor']))
        return

    def load_province(self, which='main', **kwargs):
        """
        Add a China map which contains every province.
        More arguments can be seen for self.add_shape
        """
        file = os.path.dirname(__file__) + '/shapefiles/cnhimap.shp'
        self.add_shape(file, which=which, **kwargs)

    def load_china(self, which='main', **kwargs):
        """
        Add a China map.
        More arguments can be seen for self.add_shape
        """
        file = os.path.dirname(__file__) + '/shapefiles/china_country.shp'
        self.add_shape(file, which=which, **kwargs)

    def load_river(self, which='main', **kwargs):
        """
        Add the rivers in China.
        More arguments can be seen for self.add_shape
        """
        file = os.path.dirname(__file__) + '/shapefiles/rivers.shp'
        self.add_shape(file, which=which, **kwargs)

    def load_tp(self, which='main', **kwargs):
        """
        Add the Qinghai Tibet Plateau.
        More arguments can be seen for self.add_shape
        """
        file = os.path.dirname(__file__) + '/shapefiles/TPBoundary_2500.shp'
        self.add_shape(file, which=which, **kwargs)

    def add_box(self, x, y, fill=False, **kwargs):
        """
        Add a rectangle box.
        
        **Arguments:**

        *x*
            A list that contains three or two items. The first two iterms are the
            beginning and the ending of the longitudes.

        *y*
            The same as x, but was lattitude's.
        
        More arguments can be seen for matplotlib.patches.Rectangle
        """
        if not self.MainWidget:
            self.__subplot(111)
        if 'transform' not in kwargs:
            kwargs.update({"transform":self._para['transform']})
        tmp = [x[0], y[0], x[1] - x[0], y[1] - y[0]]
        self.ax.add_patch(
            patches.Rectangle(tmp[:2], tmp[2], tmp[3], fill=fill, **kwargs))
        return

    def text(self, *args, **kwargs):
        """
        See matplotlib.pyplot.text
        """
        if not self.MainWidget:
            self.__subplot(111)
        if 'transform' not in kwargs:
            kwargs.update({"transform":self._para['transform']})
        self.ax.text(*args, **kwargs)

    def title(self, label, *args, **kwargs):
        """
        See matplotlib.pyplot.title
        """
        plt.title(label, *args, **kwargs)

    def show(self):
        if self.__need_subplot:
            self._subplot()
            self._extend(self.x, self.y)
        plt.show()

    def savefig(self, *arg, **kwargs):
        """
        See matplotlib.pyplot.savefig
        """
        plt.savefig(*arg, **kwargs)


class ncmap(map):

    def __init__(self, file, xticks=None, yticks=None, center=110, **kwargs):
        """
        Create a map object from a nc file.

         **Arguments:**

        *file*
            The nc file path or a xarray.Dataset object.
            If it is a filepath, we use xarray.open_dataset(file) to open it.
        
        More arguments can be seen for oyl.map
        """

        f = xr.open_dataset(file) if isinstance(file, str) else file
        lat = 'lat' if 'lat' in f.coords else 'latitude'
        lon = 'lon' if 'lon' in f.coords else 'longitude'
        lat, lon = f[lat].data, f[lon].data
        dx, dy = lon[1] - lon[0], lat[1] - lat[0]

        super().__init__([lon.min(), lon.max(), dx], [lat.min(), lat.max(), dy],
                         xticks=xticks, yticks=yticks, center=center, **kwargs)
        self.x = lon
        self.y = lat


class shpmap(map):

    def __init__(self, file, encoding='gbk', shape=True, xticks=None, yticks=None, center=110, **kwargs):

        """
        Create a map object from a nc file.

        **Arguments:**

        *file*
            The nc file path or a xarray.Dataset object.
            If it is a filepath, we use xarray.open_dataset(file) to open it.
        
        More arguments can be seen for oyl.map
        """

        from shapefile import Reader
        try:
            sf = Reader(file, encoding=encoding)
        except:
            sf = Reader(file, encoding='utf-8')
        pts = []
        for i, sp in enumerate(sf.shapes()):
            pts.append(np.array(sp.points))
        pts = np.vstack(pts)
        xmax, xmin = np.max(pts[:, 0]), np.min(pts[:, 0])
        ymax, ymin = np.max(pts[:, 1]), np.min(pts[:, 1])

        super().__init__([xmin, xmax], [ymin, ymax], xticks=xticks, yticks=yticks, center=center, **kwargs)
        if shape:
            self.add_shape(file, encoding=encoding)


def set_xy_delta(step=None, axis='x'):
    if step == None:
        return
    else:
        ax = plt.gca()
    if (axis != 'y') | (axis != 'Y'):
        ax.xaxis.set_major_locator(plt.MultipleLocator(step))
    else:
        ax.yaxis.set_major_locator(plt.MultipleLocator(step))


def font(font='simsun'):
    plt.rcParams['font.sans-serif'] = [font]
    plt.rcParams['axes.unicode_minus'] = False


def mat_bar(matrix, **kwargs):
    para = {'number': True, 'width': 0.2, 'legend': None, 'fmt': '.1f',
            'fontsize': 8,
            'color': ['blue', 'deepskyblue', 'orange', 'red'],
            }
    para.update(kwargs)
    data = np.array([matrix]) if matrix.ndim < 2 else np.array(matrix)
    shp = data.shape

    ax = plt.gca()
    ax.xaxis.set_major_locator(plt.MultipleLocator(1))
    for j in range(shp[0]):
        idx = j if j < len(para['color']) else j % len(para['color'])
        w = 0.5 * para['width'] - para['width'] * shp[0] / 2 + j * para['width']
        x, y = np.arange(1 + w, shp[1] + w + 1), data[j, :]

        ax.bar(x, y, width=para['width'], color=para['color'][idx])

        if para['number']:
            for a, b in zip(x, y):
                txt = '{:' + para['fmt'] + '}'
                ax.text(a - 0.5 * para['width'] + 0.01, b + 0.015 * np.max(data), txt.format(b),
                        fontsize=para['fontsize'])

        if para['legend'] is not None:
            ax.legend(para['legend'], framealpha=0.4)


if __name__ == '__main__':
    d = np.arange(31 * 21).reshape(21, 31)
    m = map([70, 130, 2], [15, 55, 2], projection=ccrs.EquidistantConic(110, 60))
    m.axes([0.1, 0.1, 0.5, 0.5])
    c = m.contourf(d, cmap=plt.cm.bwr)
    m.load_tp()
    m.load_china()
    m.axes([0.6, 0.2, 0.3, 0.3])
    m.load_china()
    c = m.contourf(d, cmap=plt.cm.bwr)
    m.show()



from setuptools import setup,find_packages
setup(name='oyl',version='2.4.0',author='Lin Ouyang', 
    packages=["oyl","oyl.nn"], 
    include_package_data=True,
    install_requires=["numpy","matplotlib","cartopy","pandas",
                      "xarray","sklearn","pyshp"]
)
